import unicodedata
from itertools import product

import pandas as pd


def generate_all_odds(df, impute_cols):
    """
    Gera um dataset com todas as combinacoes possiveis para as chaves passadas
        Args:
            df (pd.DataFrame): dataframe a ser submetido ao processo
            impute_cols (list): lista de colunas a ter todos os valores possiveis gerados

        Returns:
            imputed_df (pd.DataFrame): dataset contendo todas as combinacoes
            possiveis de chaves
    """
    base_sets = {}
    for col in impute_cols:
        base_sets[col] = df[col].unique()

    imputed_df = pd.DataFrame(
        product(*[base_sets[col] for col in impute_cols]), columns=impute_cols
    )

    return imputed_df


def strip_accents(text):
    """
    Função que devolve um texto sem acentos

    Args:
        text (str): texto com acento

    Returns:
        str: texto sem acento
    """
    text = unicodedata.normalize("NFD", text).encode("ascii", "ignore").decode("utf-8")

    return str(text)


def type_conversion(df, old_types="float64", new_type="float32"):
    """
    Função que converte todas as colunas de um tipo para outro

    Args:
      df(pd.Dataframe): Dataset de entrada
      old_type (datatype): tipo do dado de entrada
      new_type (datatype): tipo do dado de saida

    Returns:
      df(pd.Dataframe): Dataset com tipos modificados
    """
    cols = df.select_dtypes(old_types).columns
    df.loc[:, cols] = df.loc[:, cols].astype(new_type)
    return df



import re
import gc
import os
import json
import logging
import tempfile
import warnings
import pandas as pd
import numpy as np

from functools import partial
from io import BytesIO, StringIO
from urllib.request import urlretrieve
from typing import Union

from azure.storage.filedatalake import DataLakeServiceClient, DataLakeFileClient
from azure.storage.blob import BlobServiceClient, BlobClient
from azure.identity import DefaultAzureCredential
from azure.core.exceptions import ResourceNotFoundError

# Set the logging level for all azure-* libraries
logger = logging.getLogger('azure')
logger.setLevel(logging.ERROR)
warnings.filterwarnings('ignore')

DEFAULT_CREDENTIAL_KWARGS = json.loads(os.getenv('DEFAULT_CREDENTIAL_KWARGS', '{}'))
DEFAULT_SERVICE_KWARGS = json.loads(os.getenv('DEFAULT_SERVICE_KWARGS', '{}'))
DEFAULT_CONN_KWARGS = json.loads(os.getenv('DEFAULT_CONN_KWARGS', '{}'))
DEFAULT_GLOB_CONN_KWARGS = json.loads(os.getenv('DEFAULT_GLOB_CONN_KWARGS', '{}'))
DEFAULT_BLOB_SERVICE = os.getenv('DEFAULT_BLOB_SERVICE', 'gen2')
DEFAULT_UPLOAD_MODE = os.getenv('DEFAULT_UPLOAD_MODE', 'full')
DEFAULT_NUM_THREADS = os.getenv('DEFAULT_NUM_THREADS', -1)


def create_blob_service(
    uri,
    conn_type=DEFAULT_BLOB_SERVICE,
    service_kwargs=DEFAULT_SERVICE_KWARGS,
    credential_kwargs=DEFAULT_CREDENTIAL_KWARGS
):
    """
    Creates the connection service to a certain storage account
    Args:
        uri (str): Url with adress in a storage account
        conn_type (str): Type of connection to Azure. Can receive 'blob' or 'gen2'.
        service_kwargs (dict): dict with connection params to Azure services.
    Returns:
        object: DataLakeServiceClient (gen2)/BlobServiceClient (blob), depending on conn_type
    """

    credential = DefaultAzureCredential(**credential_kwargs)

    account_name = uri.split('//')[1].split('.')[0]

    if conn_type == 'gen2':
        dlService = DataLakeServiceClient(
            account_url=f"https://{account_name}.dfs.core.windows.net",
            credential=credential,
            **service_kwargs
        )

    elif conn_type == 'blob':
        dlService = BlobServiceClient(
            account_url=f"https://{account_name}.blob.core.windows.net",
            credential=credential,
            **service_kwargs
        )

    return dlService


def rename_file(uri_old, uri_new, conn_type=DEFAULT_BLOB_SERVICE):

    """
    Renames a file in an Azure Data Lake environment.
    Args:
        uri_old (str): old URL
        uri_new (str): new URL
        conn_type (str): Type of connection to Azure. Can receive 'blob' or 'gen2'.
    """

    service_client = create_blob_service(uri=uri_old, conn_type=conn_type)
    container_name = uri_old.split('/')[3]
    old_name = '/'.join(uri_old.split('/')[4:])
    new_name = '/'.join(uri_new.split('/')[4:])

    try:
        if conn_type == 'gen2':
            file_system_client = service_client.get_file_system_client(
                file_system=container_name
            )
            file_client = file_system_client.get_file_client(file_path=old_name)

        elif conn_type == 'blob':
            file_client = service_client.get_blob_client(
                container=container_name, blob=old_name
            )

        file_client.rename_file(new_name)

    finally:
        file_client.close()
        del file_client

        if conn_type == 'gen2':
            file_system_client.close()
            del file_system_client

        service_client.close()
        del service_client

        gc.collect()


def upload_chunks(
    file_client: Union[DataLakeFileClient, BlobClient],
    data: Union[BytesIO, StringIO],
    **upload_kwargs
):
    """
    Generic function to upload files in chunks

    Args:
        file_client (Union[DataLakeFileClient,BlobClient]): Azure client file.
        data (Union[BytesIO,StringIO]): Data stream to be uploaded
        **upload_kwargs: Args to be used with upload_data/upload_blob.
        See DataLakeFileClient.upload_data ou BlobFileClient/upload_blob documentation
        for details.

    """
    message = "You must pass chunk_size if using upload_mode=chunks"
    assert 'chunk_size' in upload_kwargs, message

    if 'chunk_size' in upload_kwargs:
        chunk_size = upload_kwargs.pop('chunk_size')

    upload_kwargs.pop('overwrite')

    if isinstance(file_client, DataLakeFileClient):
        file_client.create_file()

    elif isinstance(file_client, BlobClient):
        file_client.create_append_blob()

    while True:
        read_data = data.read(chunk_size)

        if not read_data:
            break

        if isinstance(file_client, DataLakeFileClient):
            if file_client.exists():
                filesize_previous = file_client.get_file_properties().size
            else:
                filesize_previous = 0

            file_client.append_data(
                data=read_data,
                offset=filesize_previous,
                length=len(read_data),
                **upload_kwargs
            )
            file_client.flush_data(filesize_previous+len(read_data))

        elif isinstance(data, StringIO):
            read_data = ''.join(read_data)

        elif isinstance(file_client, BlobClient):
            file_client.append_block(read_data, **upload_kwargs)


def upload_data(
    byte_stream: Union[BytesIO, StringIO],
    file_client: Union[DataLakeFileClient, BlobClient],
    upload_mode: str = DEFAULT_UPLOAD_MODE,
    **upload_kwargs
):
    """
    Generic function to upload files.

    Args:
        byte_stream (Union[BytesIO,StringIO]): Data stream to be uploaded
        file_client (Union[DataLakeFileClient,BlobClient]): Azure client file.
        upload_mode (str): 'full' for direct upload or 'chunks' for uploading in parts.
            If 'chunks' is set. 'chunk_size' has to be declared in 'upload_kwargs'.
        **upload_kwargs: Args to be used with upload_data/upload_blob.
            See DataLakeFileClient.upload_data ou BlobFileClient/upload_blob documentation
        for details.
    """
    if isinstance(file_client, DataLakeFileClient):
        delete_func = file_client.delete_file
    elif isinstance(file_client, BlobClient):
        delete_func = file_client.delete_blob

    if upload_mode == 'full':
        if isinstance(file_client, DataLakeFileClient):
            upload_func = file_client.upload_data

        elif isinstance(file_client, BlobClient):
            if 'chunk_size' in upload_kwargs:
                upload_kwargs.pop('chunk_size')
            upload_func = file_client.upload_blob

    elif upload_mode == 'chunks':
        upload_func = partial(upload_chunks, file_client=file_client)

    if file_client.exists():
        delete_func()

    byte_stream.seek(0)

    if upload_mode == 'full':
        if isinstance(byte_stream, StringIO):
            byte_stream = ''.join(byte_stream.readlines())

    upload_func(
        data=byte_stream,
        overwrite=True,
        **upload_kwargs
    )

    return file_client


def to_any(
    byte_stream,
    uri,
    conn_type=DEFAULT_BLOB_SERVICE,
    upload_mode=DEFAULT_UPLOAD_MODE,
    verbose=1,
    **upload_kwargs
):
    """
    Funcao generica de escrita de arquivo na Azure.

    Args:
        byte_stream (stream): Stream de dados a serem subidos
        uri (url): Url a ser subida o stream de dados
        upload_mode (str): Modo de upload.'full' (de uma vez só) ou 'chunks' (em pedaços).
            No modo chunks, precisa explicitar o chunk_size no upload_kwargs.
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        **upload_kwargs: Argumentos a serem passados para a upload_data/upload_blob.
            Mais detalhes em DataLakeFileClient.upload_data ou BlobFileClient.upload_blob
    """

    service_client = create_blob_service(uri=uri, conn_type=conn_type)
    container_name = uri.split('/')[3]
    blob_name = '/'.join(uri.split('/')[4:])

    if verbose > 0:
        logger.info(f'Writing {blob_name}')

    if conn_type == 'gen2':
        file_system_client = service_client.get_file_system_client(
            file_system=container_name
        )
        file_client = file_system_client.get_file_client(file_path=blob_name)

    elif conn_type == 'blob':
        file_client = service_client.get_blob_client(
            container=container_name, blob=blob_name
        )

    try:
        file_client = upload_data(
            byte_stream=byte_stream,
            file_client=file_client,
            upload_mode=upload_mode,
            **upload_kwargs
        )

    finally:
        file_client.close()
        del file_client

        if conn_type == 'gen2':
            file_system_client.close()
            del file_system_client

        service_client.close()
        del service_client

        gc.collect()


def read_any(uri, func, conn_type=DEFAULT_BLOB_SERVICE, **kwargs):
    """
    Função de download genérico de arquivo na Azure

    Args:
        uri (url): Url a ser subida o stream de dados
        func: Função de leitura do arquivo
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        **kwargs: Argumentos a serem passados para a função de leitura

    Returns:
        Output da função func
    """

    def close_connections(*connections):
        for conn in connections:
            if conn:
                conn.close()
                del conn

    service_client = create_blob_service(uri, conn_type=conn_type)
    container_name = uri.split('/')[3]
    blob_name = '/'.join(uri.split('/')[4:])

    byte_stream = BytesIO()

    if conn_type == 'gen2':
        file_system_client = service_client.get_file_system_client(
            file_system=container_name
        )
        file_client = file_system_client.get_file_client(blob_name)

    elif conn_type == 'blob':
        file_client = service_client.get_blob_client(
            container=container_name, blob=blob_name
        )

    assert file_client.exists(), f'Could not find blob in {blob_name}'

    try:
        if conn_type == 'gen2':
            byte_stream.write(file_client.download_file().readall())
        elif conn_type == 'blob':
            byte_stream.write(file_client.download_blob().readall())

        byte_stream.seek(0)
        df = func(byte_stream, **kwargs)

    except Exception as e:
        raise Exception(f'Could not read blob in {blob_name}: {e}')

    finally:
        close_connections(byte_stream, file_client, file_system_client, service_client)
        gc.collect()

    return df


def to_parquet(
    df,
    uri,
    conn_type=DEFAULT_BLOB_SERVICE,
    upload_kwargs=DEFAULT_CONN_KWARGS,
    upload_mode=DEFAULT_UPLOAD_MODE,
    **kwargs
):
    """
    Função de escrita para arquivos parquet e consequente subida a Azure.

    Args:
        df (pd.DataFrame): DataFrame a se escrever em parquet e subir a Azure
        uri (str): String com url a ser escrito o arquivo.
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        upload_kwargs (dict): Argumentos a serem passados para a upload_data/upload_blob.
            Mais detalhes em DataLakeFileClient.upload_data ou BlobFileClient.upload_blob
        **kwargs: Argumentos a serem passados para a função de escrita em parquet.
            Consultar df.to_parquet para mais detalhes.
    """
    byte_stream = BytesIO()
    df.to_parquet(byte_stream, use_deprecated_int96_timestamps=True, **kwargs)
    to_any(byte_stream, uri, conn_type=conn_type, upload_mode=upload_mode, **upload_kwargs)


def to_excel(
    df,
    uri,
    conn_type=DEFAULT_BLOB_SERVICE,
    mode='pandas',
    upload_kwargs=DEFAULT_CONN_KWARGS,
    upload_mode=DEFAULT_UPLOAD_MODE,
    **kwargs
):
    """
    Função de escrita para arquivos excel e consequente subida a Azure.

    Args:
        df (pd.DataFrame): DataFrame a se escrever em excel e subir a Azure
        uri (str): String com url a ser escrito o arquivo.
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        mode (str): Modo de escrita de excel. No momento somente suporte a 'pandas'.
        upload_kwargs (dict): Argumentos a serem passados para a upload_data/upload_blob.
            Mais detalhes em DataLakeFileClient.upload_data ou BlobFileClient.upload_blob
        **kwargs: Argumentos a serem passados para a função de escrita em parquet.
            Consultar df.to_parquet para mais detalhes.
    """

    # Pyexcelerate still not supported
    if mode == 'pyexcelerate':
        logger.warn('to_excel method is currently not supported with pyexcelerate mode')
        mode = 'pandas'

    func_dict = {'pandas': pd.DataFrame.to_excel}

    byte_stream = BytesIO()
    func_dict[mode](df, byte_stream, **kwargs)
    to_any(
        byte_stream,
        uri,
        conn_type=conn_type,
        upload_mode=upload_mode,
        **upload_kwargs
    )


def to_csv(
    df,
    uri,
    conn_type='gen2',
    encoding='utf-8',
    upload_kwargs=DEFAULT_CONN_KWARGS,
    upload_mode=DEFAULT_UPLOAD_MODE,
    **kwargs
):
    """
    Função de escrita para arquivos csv e consequente subida a Azure.

    Args:
        df (pd.DataFrame): DataFrame a se escrever em csv e subir a Azure
        uri (str): String com url a ser escrito o arquivo.
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        encoding (str): String com encoding do arquivo a ser subido.
        upload_kwargs (dict): Argumentos a serem passados para a upload_data/upload_blob.
            Mais detalhes em DataLakeFileClient.upload_data ou BlobFileClient.upload_blob
        **kwargs: Argumentos a serem passados para a função de escrita em parquet.
            Consultar df.to_parquet para mais detalhes.
    """
    # Csv writing currently not supported in blob conn_type
    if conn_type == 'blob':
        logger.warn('to_csv method is currently not supported with blob conn_type')
        conn_type = 'gen2'

    byte_stream = StringIO()
    df.to_csv(byte_stream, encoding=encoding, **kwargs)

    to_any(
        byte_stream,
        uri,
        encoding=encoding,
        conn_type=conn_type,
        upload_mode=upload_mode,
        **upload_kwargs
    )


def build_re(glob_str):
    """
    Função que reimplementa o fnmatch.translate.
    Args:
        glob_str (str): String com glob pattern.
    Returns:
        str: String traduzida para regex pattern.
    """

    opts = re.compile('([.]|[*][*]/|[*]|[?])|(.)')
    out = ''
    for (pattern_match, literal_text) in opts.findall(glob_str):
        if pattern_match == '.':
            out += '[.]'
        elif pattern_match == '**/':
            out += '(?:.*/)?'
        elif pattern_match == '*':
            out += '[^/]*'
        elif pattern_match == '?':
            out += '.'
        elif literal_text:
            out += literal_text
    return out


def glob(uri, conn_kwargs=DEFAULT_GLOB_CONN_KWARGS, **kwargs):
    """
    Função que permite, dado uma url, pegar todos os endereços de arquivos da pasta que
    atendam aos requisitos.

    Args:
        uri (str): Url com padrão fnmatch de regex.
            Já possuimos alguns recursos com suporte:
                - * permite pegar qualquer url que possua qualquer string no lugar de *
                - (word1|word2) permite filtrar urls que contenham as palavras word1 ou word2
                - dentre outros
        conn_kwargs (dict): Dicionário com argumentos de conexão a serem passados para a
            função de get_paths
        **kwargs (dict): Argumentos da função get_paths
    Returns:
        list: Lista com url dos diretórios que atendem ao requisito da uri
    """

    blob_service = create_blob_service(uri, conn_type='gen2')
    container_name = uri.split('/')[3]
    container_url = '/'.join(uri.split('/')[:4])
    blob_name = '/'.join(uri.split('/')[4:])
    container_client = blob_service.get_file_system_client(file_system=container_name)

    # Como a get_paths exige que a path exista, nós quebramos a url em duas variaveis:
    # - path: com a parte da url que exista uma pasta
    # - path_suffix: com a query desejada dentro dessa pasta

    # TODO: avaliar as sequencias de regex desse codigo
    regex = re.compile("^(.*?[\*,\(,\[])")
    lista = regex.split(blob_name)

    if len(lista) == 1:
        path = lista[0]
        path_suffix = ''
    else:
        new_split = '/'.join(lista[:-1]).split('/')
        path = '/'.join(new_split[:-1])
        path_suffix = new_split[-1] + lista[-1]

    list_blobs = [
        container_url + '/' + unit.name
        for unit in container_client.get_paths(path=path, **kwargs, **conn_kwargs)
    ]

    result_list = []

    if len(list_blobs) == 0:
        print('No file match the specified criteria')
        return result_list

    if len(path_suffix) == 0:
        return list_blobs

    else:
        path_suffix = build_re(path_suffix)  # traduz fnmatch para regex
        result_list = [
            i for i in np.array(list_blobs) if re.search(path_suffix, i) is not None
        ]

    return result_list


def read_parquet(uri, mode='pandas', conn_type=DEFAULT_BLOB_SERVICE, **kwargs):

    """
    Função de leitura genérica de parquet
    Args:
        uri (str): Url do arquivo
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        **kwargs: Argumentos extras das funções de leitura disponiveis.
    Returns:
        pd.DataFrame: DataFrame desejado
    """
    func = {'pandas': pd.read_parquet}
    func = func[mode]
    return read_any(uri, func, conn_type=conn_type, **kwargs)


def read_csv(uri, mode='pandas', conn_type=DEFAULT_BLOB_SERVICE, **kwargs):

    """
    Função de leitura genérica de csv
    Args:
        uri (str): Url do arquivo
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        **kwargs: Argumentos extras das funções de leitura disponiveis.
    Returns:
        pd.DataFrame: DataFrame desejado
    """
    func = {'pandas': pd.read_csv}
    func = func[mode]
    return read_any(uri, func, conn_type=conn_type, **kwargs)


def read_excel(uri, mode='pandas', conn_type=DEFAULT_BLOB_SERVICE, **kwargs):

    """
    Função de leitura genérica de excel
    Args:
        uri (str): Url do arquivo
        mode (str): Pode ser 'pandas' para leitura padrão pelo pandas.
        conn_type (str): String com serviço de conexão a Azure. Pode ser blob ou gen2.
        **kwargs: Argumentos extras das funções de leitura disponiveis.
    Returns:
        pd.DataFrame: DataFrame desejado
    """

    # A partir de uma determinada versao, o xlrd parou de dar suporte a xlsx.
    # Usa-se por padrão a engine openpyxl. Se ela não for passada, agnt força a engine
    if ('.xlsx' in uri) & ('engine' not in kwargs):
        kwargs['engine'] = 'openpyxl'

    func = {'pandas': pd.read_excel}
    func = func[mode]
    return read_any(uri, func, conn_type=conn_type, **kwargs)


def read_url(uri, sas_token, _format, **kwargs):
    """Read from a container with SAS token """
    with tempfile.NamedTemporaryFile() as tf:
        url_tok = uri + sas_token
        urlretrieve(url_tok, tf.name)
        df = read_any(uri=tf.name, _format=_format, **kwargs)
        return df


def file_exists(path):
    """ Checa se um arquivo de Data Lake Azure existe """
    last_dir = path.replace(path.split('/')[-1], "*")

    try:
        if path in glob(last_dir):
            return True
        else:
            return False
    except ResourceNotFoundError:
        return False



import logging
import os

from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient

log = logging.getLogger(__name__)

# Set the logging level for all azure-* libraries
logger = logging.getLogger('azure')
logger.setLevel(logging.ERROR)


def load_from_keyvault():
    """Load secrets from Azure KeyVault into the environment"""

    kv_url = os.getenv("AZURE_KEYVAULT_URL")
    if not kv_url:
        log.warning("KeyVault URL not set, make sure to fix this!")
        return

    log.info("Connecting to Azure KeyVault: " + kv_url)
    creds = DefaultAzureCredential(exclude_shared_token_cache_credential=True)
    client = SecretClient(kv_url, creds)

    log.info("Loading secrets into memory")
    for secret in client.list_properties_of_secrets():
        name = secret.name.replace("-", "_")
        if name in os.environ:
            log.info(f"Secret '{name}' already in environment, not overwriting")
        else:
            log.info("Loading secret: " + name)
            os.environ[name] = client.get_secret(secret.name).value



import pandas as pd

from datetime import datetime


def is_weekend(date, include_friday=False):
    """
    Retorna uma resposta se o dia especificado eh fim de semana

    Args:
        date (datetime.date): data a ser testada
        include_friday (bool): True para incluir sexta no teste logico

    Returns:
        bool: True se a data passada esta no final de semana
    """
    if include_friday:
        weekend = [4, 5, 6]
    else:
        weekend = [5, 6]

    return date.weekday() in weekend


def week_to_date_row(week, ref=pd.to_datetime("1989-12-31")):
    """
    Função que retorna a data do primeiro dia da semana
    Para nós, o primeiro dia é sempre domingo

    Args:
       week (str): Entidade com os números da semana
       ref (str): Data referencia de inicio

    Returns:
       str: Data acrescida dos números da semana
    """

    return ref + datetime.timedelta(days=week * 7)


def week_to_date_column(week_column, ref=pd.to_datetime("1989-12-31")):
    """
    Função que retorna série com a data do primeiro dia da semana.
    Para nós, o primeiro dia é sempre domingo

    Args:
       week_column (pd.Series): Coluna com os números da semana
       ref (str): Data referencia de inicio

    Returns:
       pd.Series: Coluna com o primeiro dia da semana
    """
    return pd.to_timedelta(week_column * 7, 'd') + ref



import pandas as pd
from joblib import Parallel, delayed


def applyParallel(
    dfGrouped,
    func,
    n_jobs=-1,
    concat_results=True,
    backend='loky',
    plot_mode=False,
    verbose_par=0,
    *args,
    **kwargs
):
    """
    Função que implementa o apply do pandas usando backend paralelizavel.

    Args:
        dfGrouped (pd.DataFrame or zip): Objeto groupby do pandas
            ou zip(objeto groupby + axes do matplotlib) para plot_mode=True
        func: Funcao definida pelo usuario
        n_jobs: Numero de jobs a serem utilizados no calculo da funcao
        concat_results (bool): Se o usuario deseja que os resultados sejam concatenados
            e retornados num novo dataframe
        backend (str): Backend paralelo. Pode ser 'loky', 'threading', 'multiprocessing'.
        plot_mode (bool): Se esta usando em modo de plot ou nao
        verbose (int): Parametro que controla verbosidade do Parallel
            (o quao ele vai printar mais)
        *args: Argumentos da funcao a ser aplicada
        **kwargs: Argumentos de palavra-chave da funcao a ser aplicada

    Returns:
        Se concat_results=True: pd.DataFrame: DataFrame com resultados calculados.
        Se False: dict: Dicionário com chave sendo o grupo e valor sendo o resultado
        da função do grupo especifico.
    """

    def temp_func(func, name, group, *args, **kwargs):
        return name, func(group, *args, **kwargs)

    if plot_mode:
        backend = 'threading'
        series_par = dict(
            Parallel(n_jobs=n_jobs, backend=backend, verbose=verbose_par)(
                delayed(temp_func)(func, name, group, ax=ax, *args, **kwargs)
                for (name, group), ax in dfGrouped
            )
        )

    else:
        series_par = dict(
            Parallel(n_jobs=n_jobs, backend=backend, verbose=verbose_par)(
                delayed(temp_func)(func, name, group, *args, **kwargs)
                for name, group in dfGrouped
            )
        )

    if concat_results:
        return pd.concat(series_par, sort=True)
    else:
        return series_par



import numpy as np
import pandas as pd


def create_lag_features(df, group_columns, target_column, num_lags, start=1, step=1):
    """
        Args:
            df (pd.DataFrame): dataset
            group_columns (list): colunas a serem agrupadas
            target_column (str): coluna a ser aplicada a funcao
            num_lags (int): numero maximo de lag
            start (int): numero inicial de lag
            step (int): intervalo entre os numeros de lag

        Returns:
            df (pd.DataFrame): dataset com colunas de lag
    """
    for i in range(start, num_lags + 1, step):
        column_name = f"{target_column}_lag_{i}"
        df[column_name] = df.groupby(group_columns)[target_column].shift(i)

    return df


def create_ma_features(df, group_columns, target_column, num_periods):
    """
      Args:
          df (pd.DataFrame): dataset
          group_columns (list): colunas a serem agrupadas
          target_column (str): coluna a ser aplicada a funcao
          num_periods (int): numero minimo de janelas a serem usadas para a media movel

      Returns:
          df (pd.DataFrame): dataset com colunas de ma
  """

    column_name = f"{target_column}_ma_{num_periods}"

    df[column_name] = (
        df.groupby(group_columns)[target_column]
        .rolling(num_periods)
        .mean()
        .shift(1)
        .reset_index(np.arange(len(group_columns)).tolist(), drop=True)
    )

    return df


def month_encoder(month, ref=pd.to_datetime("1990-01-01")):
    """
    Função que retorna o número de meses que passaram-se de uma data de referência

    Args:
      month (pd.Series): Série com datas
      ref (str): Data referencia de inicio

    Returns:
      PeriodArray: Série com número de meses que se passaram desde a data de referência
    """
    return (month.to_period("M") - ref.to_period("M")).n


def new_date_encoder(date: pd.Series,
                     freq: str = 'M',
                     date_start: str = '1970-01-01'):

    """
    Função que retorna uma série com o número do mês/dia relativo
        a data de referência

    Args:
        date (pd.Series): Série com datas
        freq (str): Permite em meses ('M') e em dias ('D')
        date_start (str): Data referencia de inicio

    Returns:
        pd.Series: Série com número do mês/dia relativo a data de referência
    """

    date = pd.DatetimeIndex(date)

    if freq == 'M':
        date = pd.to_datetime(
            np.where(date.day == 1, date, date - pd.offsets.MonthBegin(n=1)),
            format='%Y-%m-%d'
        )

    date_start = pd.DatetimeIndex(np.full_like(date, date_start))

    return date.to_period(freq).astype('int') - date_start.to_period(freq).astype('int')


def week_encoder(date, ref=pd.to_datetime("1989-12-31")):
    """
    Função que retorna o número de semanas que passara-se
    desde uma data de referência (começando por domingo)

    Args:
      date (pd.Series): Série com datas
      ref (str): Data referencia de inicio

    Returns:
      pd.Series: Série com número de semanas passadas após a data referência
    """
    return np.floor((date - ref).dt.days / 7)



import re
import pandas as pd


def path_join(*args, sep='/'):
    """ Une partes de um path """
    return f"{sep}".join([*args])


def get_date_from_names(names_list, sep="_", occ=-1):
    """ Given a list with filenames, returns the dates contained within each filename
    """
    result_list = []
    for item in names_list:
        if sep == "_":
            match = re.search(r"\d{4}_\d{2}_\d{2}", item.split("/")[occ])
        elif sep == "-":
            match = re.search(r"\d{4}-\d{2}-\d{2}", item.split("/")[occ])
        elif sep == "/":
            match = re.search(r"\d{4}/\d{2}/\d{2}", item)
        elif sep == "":
            match = re.search(r"\d{4}\d{2}\d{2}", item.split("/")[occ])

        result_list.append(str(match.group()))

    if sep == "-":
        return pd.to_datetime(result_list, format="%Y-%m-%d")
    elif sep == "/":
        return pd.to_datetime(result_list, format="%Y/%m/%d")
    elif sep == "_":
        return pd.to_datetime(result_list, format="%Y_%m_%d")
    elif sep == "":
        return pd.to_datetime(result_list, format="%Y%m%d")
    else:
        raise "Date format not supported"



import logging
import numpy as np
import pandas as pd
import re

from paeio.io import io
from paeio.data_cleaning.path import get_date_from_names
from paeio.data_cleaning.parallelism import applyParallel

logger = logging.getLogger(__name__)


def load_history(
    glob, initial_date='2019-01-01', sep="-", _format='parquet', columns=None, occ=-1
):
    """Le o dataset de cada mes entre a data inicial e hj e junta em um unico arquivo

    Args:
        initial_date (str): Data minima para estar no output. Default: None
        ext (str): Tipo de extensao do arquivo. Default: parquet
    """

    all_files = np.array(io.glob(glob))
    all_dates = get_date_from_names(all_files, sep=sep, occ=occ)

    index = all_dates >= initial_date
    files_to_load = all_files[index]

    if _format == 'parquet':
        all_df = pd.concat(
            [io.read_parquet(f, columns=columns) for f in files_to_load],
            sort=True,
        )
    elif _format == 'csv':
        all_df = pd.concat(
            [io.read_csv(f, sep=';', decimal=',') for f in files_to_load],
            sort=True,
        )
    else:
        all_df = pd.concat([io.read_excel(f) for f in files_to_load], sort=True)
    return all_df


def read_whole_path(uri, columns=None):
    """ Le todos os arquivos de um path """
    final_df = pd.DataFrame()
    for i in io.glob(uri):
        df = io.read_parquet(i, columns=columns)
        final_df = final_df.append(df)

    return final_df


def load_last_file(glob, _format='parquet'):
    """Le o ultimo arquivo de um historico de arquivos

    Args:
        glob (str): path
        ext (str): Tipo de extensao do arquivo. Default: .parquet
    """
    file_to_load = io.glob(glob)[-1]

    if _format == 'parquet':
        df = io.read_parquet(file_to_load)
    elif _format == 'csv':
        df = io.read_csv(file_to_load, sep=';', decimal=',')
    else:
        df = io.read_excel(file_to_load)
    return df


def extract_file(
    x,
    file_format,
    verbose=0,
    file_func=lambda x: x,
    keep_origin_col=False,
    errors='raise',
    mode='read',
    **kwargs
):

    if verbose > 0:
        if mode == 'read':
            print(f"Reading {x['directory'].values[0]}")
        elif mode == 'rename':
            print(f"Renaming {x['directory'].values[0]}")

    if mode == 'read':
        dict = {'parquet': io.read_parquet, 'excel': io.read_excel, 'csv': io.read_csv}
    elif mode == 'rename':
        dict = {form: io.rename_file for form in ['parquet', 'excel', 'csv']}

    if file_format in list(dict.keys()):
        func = dict[file_format]
        try:
            df = func(x['directory'].values[0], **kwargs)

            if keep_origin_col:
                df['origin'] = x['directory'].values[0]

            df = file_func(df)

        except Exception as e:
            if errors == 'raise':
                raise e
            elif errors == 'ignore':
                return pd.DataFrame()

    else:
        try:
            df = file_func(x['directory'].values[0], **kwargs)

        except Exception as e:
            if errors == 'raise':
                raise e
            elif errors == 'ignore':
                return pd.DataFrame()

    return df


def datalake_walk(
    base_folder,
    file_format='parquet',
    min_date='2000-01-01',
    last_layer='folder',
    max_date='2099-12-31',
    verbose=0,
    n_jobs=-1,
    file_func=lambda x: x,
    date_sep='-',
    occ=-1,
    keep_origin_col=False,
    errors='raise',
    mode='read',
    **kwargs
):
    """
    Função que permite navegar de forma recursiva dentro de um sistema de pastas
    dividido em anos, meses e dias e concatena os arquivos navegados.
    Ele permite implementação de lógica incremental a partir do parâmetro min_date.

    Args:
        base_folder (str): Diretório raiz que contem os anos, meses e dias
        min_date (str): Data mínima desejada
        file_format (str): Tipo de arquivo a ser lido na pasta.
            Pode ser parquet, csv ou excel.
        last_layer (str): Ultima camada de pastas. Pode ser year, month ou day.
        max_date (str): Data máxima desejada
        verbose (int): Nivel de verbosidade do codigo.
            Acima de zero, printa entradas nas pastas
        n_jobs (int): Numero de jobs maximo a ser usado na paralelização da leitura
        file_func (func): Função a ser aplicada no momento da leitura
        date_sep (str): Separador de data em pastas com arquivos com datas no nome
        occ (int): Ocorrencia do split que a data se encontra em caso de folder
        keep_origin_col (bool): Manter ou não o nome da coluna de diretorio de origem
        errors (str): Pode ser 'raise' (quebra codigo quando da erro)
            ou 'ignore' (segue com dataframe vazio)
        mode (str): Modo de leitura. Pode ser 'read' para ler o arquivo
            ou 'rename' para renomear o arquivo
        **kwargs: Argumentos a serem passados na read_any (ex: sep do read_csv)

    Returns:
        pd.DataFrame: DataFrame concatenado final com todos os arquivos que foram navegados

    """
    min_date = pd.to_datetime(min_date)
    max_date = pd.to_datetime(max_date)

    extension = file_format if file_format != 'excel' else 'xls'

    list_folders = [col for col in io.glob(base_folder) if f'.{extension}' in col]

    start = -2
    date_format = '%Y'

    if last_layer == 'day':
        start = -4
        date_format = '%Y/%m/%d'
    elif last_layer == 'month':
        start = -3
        date_format = '%Y/%m'

    df_folders = pd.DataFrame(list_folders, columns=['directory'])
    df_folders['file_name'] = df_folders['directory'].str.split('/').str.get(-1)

    if last_layer != 'folder':
        df_folders['date'] = pd.to_datetime(
            df_folders['directory'].str.split('/')
            .str.slice(start=start, stop=-1).str.join('/'),
            format=date_format
        )
        df_folders = df_folders.loc[df_folders['date'].between(min_date, max_date)]

    else:
        try:
            df_folders['date'] = get_date_from_names(
                df_folders['directory'], sep=date_sep, occ=occ
            )
            df_folders = df_folders.loc[df_folders['date'].between(min_date, max_date)]
        except:
            pass

    df = None

    if len(df_folders) == 1:
        df = extract_file(
            df_folders,
            file_format=file_format,
            verbose=verbose,
            file_func=file_func,
            keep_origin_col=keep_origin_col,
            errors=errors,
            mode=mode,
            **kwargs
        )
    elif len(df_folders) > 1:
        df = applyParallel(
            df_folders.groupby(['directory']),
            extract_file,
            keep_origin_col=keep_origin_col,
            file_format=file_format,
            file_func=file_func,
            verbose=verbose,
            errors=errors,
            n_jobs=n_jobs,
            mode=mode,
            **kwargs
        ).reset_index(drop=True)

    return df


def datalake_get_last(
    base_folder,
    file_format,
    min_date='1900-01-01',
    last_layer='folder',
    max_date='2099-12-31',
    verbose=0,
    file_func=lambda x: x,
    date_sep='-',
    occ=-1,
    keep_origin_col=False,
    n_jobs=-1,
    errors='raise',
    mode='read',
    **kwargs
):
    """
    Função que permite navegar de forma recursiva dentro de um sistema de pastas
    e pega o ultimo arquivo.

    Args:
        base_folder (str): Diretório raiz que contem os anos, meses e dias
        min_date (str): Data mínima desejada
        file_format (str): Tipo de arquivo a ser lido na pasta.
            Pode ser parquet, csv ou excel.
        last_layer (str): Ultima camada de pastas. Pode ser year, month ou day.
        max_date (str): Data máxima desejada
        verbose (int): Nivel de verbosidade do codigo.
            Acima de zero, printa entradas nas pastas
        file_func (func): Função a ser aplicada no momento da leitura
        date_sep (str): Separador de data em pastas com arquivos com datas no nome
        occ (int): Ocorrencia do split que a data se encontra em caso de folder
        n_jobs (int): Numero de jobs maximo a ser usado na paralelização da leitura
        keep_origin_col (bool): Manter ou não o nome da coluna de diretorio de origem
        errors (str): Pode ser 'raise' (quebra codigo quando da erro)
            ou 'ignore' (segue com dataframe vazio)
        mode (str): Modo de leitura. Pode ser 'read' para ler o arquivo
            ou 'rename' para renomear o arquivo
        **kwargs: Argumentos a serem passados na read_any (ex: sep do read_csv)

    Returns:
        pd.DataFrame: DataFrame concatenado final com todos os arquivos que foram navegados

    """

    min_date = pd.to_datetime(min_date)
    max_date = pd.to_datetime(max_date)

    mod_base_folder = base_folder

    regex = re.compile('\(|\{')
    group_indexes = [
        ind for ind, i in enumerate(mod_base_folder.split('/'))
        if regex.match(i) is not None
    ]

    extension = file_format if file_format != 'excel' else 'xls'

    list_folders = [col for col in io.glob(mod_base_folder) if f'.{extension}' in col]

    start = -2
    date_format = '%Y'

    if last_layer == 'day':
        start = -4
        date_format = '%Y/%m/%d'
    elif last_layer == 'month':
        start = -3
        date_format = '%Y/%m'

    df_folders = pd.DataFrame(list_folders, columns=['directory'])
    df_folders['file_name'] = df_folders['directory'].str.split('/').str.get(-1)

    if last_layer != 'folder':
        df_folders['date'] = pd.to_datetime(
            df_folders['directory'].str.split('/')
            .str.slice(start=start, stop=-1).str.join('/'),
            format=date_format
        )
        df_folders = df_folders.loc[df_folders['date'].between(min_date, max_date)]

    else:
        try:
            df_folders['date'] = get_date_from_names(
                df_folders['directory'], sep=date_sep, occ=occ
            )
            df_folders = df_folders.loc[df_folders['date'].between(min_date, max_date)]
        except:
            pass

    df_folders['group_base'] = 1

    for col in group_indexes:
        df_folders[f'group_{col}'] = df_folders['directory'].str.split('/').str.get(col)

    key = ['group_base'] + [f'group_{col}' for col in group_indexes]

    df_folders = df_folders.sort_values('directory')
    df_folders = df_folders.groupby(key, as_index=False).tail(1)

    df = None

    if len(df_folders) == 1:
        df = extract_file(
            df_folders,
            file_format=file_format,
            mode=mode,
            verbose=verbose,
            file_func=file_func,
            keep_origin_col=keep_origin_col,
            errors=errors,
            **kwargs
        )

    if len(df_folders) > 1:
        df = applyParallel(
            df_folders.groupby(['directory']),
            extract_file,
            keep_origin_col=keep_origin_col,
            file_format=file_format,
            file_func=file_func,
            verbose=verbose,
            errors=errors,
            mode=mode,
            n_jobs=n_jobs,
            **kwargs
        ).reset_index(drop=True)

    return df
